/*
    module/acl8112.c
    hardware driver for AdSys ACL-8112

    Based on David Schleef's driver for PC-LabCard PCL-711

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	Dave Andruczyk <xxx@tech.buffalostate.edu> also wrote a
	driver for the PCL-711.  I used a few ideas from his driver
	here.  His driver also has more comments, if you are
	interested in understanding how this driver works.

	The ACL-8112 driver was hacked from the sources of the PCL-711
	driver (the 744 chip used on the 8112 is almost the same as
	the 711b chip, but it has more I/O channels) by
	Janne Jalkanen (jalkanen@cs.hut.fi) and
	Erik Bunn (ebu@cs.hut.fi).

	This driver supports both TRIGNOW and TRIGCLK,
	but does not yet support DMA transfers.  It also supports
	both high (HG) and low (DG) versions of the card, though
	the HG version has been untested.  The driver defaults to
	the DG version.
*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>


/*
 *  The defines are valid also for the ACL-8112 card,
 *  as it is almost register level compatible to 711.
 */

#define PCL711_SIZE 16

#define PCL711_CTR0 0
#define PCL711_CTR1 1
#define PCL711_CTR2 2
#define PCL711_CTRCTL 3
#define PCL711_AD_LO 4
#define PCL711_DA_LO 4
#define PCL711_AD_HI 5
#define PCL711_DA_HI 5
#define PCL711_DI_LO 6
#define PCL711_DI_HI 7
#define PCL711_CLRINTR 8
#define PCL711_GAIN 9
#define PCL711_MUX 10
#define PCL711_MODE 11
#define PCL711_SOFTTRIG 12
#define PCL711_DO_LO 13
#define PCL711_DO_HI 14

typedef int bool;

/*
 * flags
 */

#define PCL711_TIMEOUT 100
#define PCL711_DRDY 0x10

comedi_devinit acl8112_init;
static int acl8112_rem(comedi_device *dev);

typedef struct{
        bool is_dg;
	int adchan;
	int ntrig;
	int aip[16];
	int last_do;
}acl8112_private;
#define devpriv ((acl8112_private *)dev->private)

static void acl8112_interrupt(int irq,void *d,struct pt_regs *regs)
{
	int lo,hi;
	comedi_device *dev=d;
	comedi_sample pt;

	hi=inb(dev->iobase+PCL711_AD_HI);
	lo=inb(dev->iobase+PCL711_AD_LO);
	outb(0,dev->iobase+PCL711_CLRINTR);
	
	pt.data=(hi<<8)|lo;
	pt.chan=devpriv->adchan;
	report_sample(dev,&pt);
	
	if(!(--devpriv->ntrig)){
		outb(1,dev->iobase+PCL711_MODE);
		dev->busy[0]=0;
		
		wake_up_interruptible(&dev->rdwait);
	} 
}
	

static int acl8112_iai(comedi_device *dev,comedi_trig *it)
{
	int hi,lo,i,mask;

	if(dev->busy[0])return -EBUSY;
	outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
	
	/* 
	 *  Set the correct channel.  The two channel banks are switched
	 *  using the mask value.
	 *  NB: To use differential channels, you should use mask = 0x30,
	 *  but I haven't written the support for this yet. /JJ
	 */

	if( it->chan >= 8 ) mask = 0x20;
	else mask = 0x10;

	outb((mask | it->chan),dev->iobase+PCL711_MUX);

	/*
	 *  Write the correct mode (software polling) and start polling by writing
	 *  to the trigger register 
	 */

	outb(1,dev->iobase+PCL711_MODE);
	outb(0,dev->iobase+PCL711_SOFTTRIG);

	for(i=0;i<PCL711_TIMEOUT;i++){
		hi=inb(dev->iobase+PCL711_AD_HI);
		if( !(hi&PCL711_DRDY) )
			break;
		udelay(25);
	}
	if(i==PCL711_TIMEOUT){
		printk("comedi%d: acl8112: A/D timeout\n",dev->minor);
		return -ETIME;
	}
	lo=inb(dev->iobase+PCL711_AD_LO);

	it->data=((hi&0xf)<<8)|lo;
		
	return 0;
}

static int acl8112_ai(comedi_device *dev,comedi_trig *it)
{
	if(dev->busy[0])return -EBUSY;
	switch(it->trigsrc){
	case COMEDI_TRIGEXT:
	        dev->busy[0]=1;
		devpriv->ntrig=it->n;
		devpriv->adchan=it->chan;
		outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
		outb(it->chan,dev->iobase+PCL711_MUX);
		/*
		 *  Set mode to "no internal trigger"
		 */
		outb(0,dev->iobase+PCL711_MODE);
		
		break;

	case COMEDI_TRIGCLK:
		dev->busy[0]=1;
		devpriv->ntrig=it->n;
		devpriv->adchan=it->chan;
		outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
		outb(it->chan,dev->iobase+PCL711_MUX);

		/*
		 *  Set timers
		 *  0x74 = Select Counter 1 | LSB/MSB | Mode=2 | Binary
		 *        Mode 2 = Rate generator
		 *
		 *  0xb4 = Select Counter 2 | LSB/MSB | Mode=2 | Binary
		 */

		outb(0x74,dev->iobase+PCL711_CTRCTL);
		outb((it->trigvar>>16)&0xff,dev->iobase+PCL711_CTR1);
		outb((it->trigvar>>24)&0xff,dev->iobase+PCL711_CTR1);
		outb(0xb4,dev->iobase+PCL711_CTRCTL);
		outb(it->trigvar&0xff,dev->iobase+PCL711_CTR2);
		outb((it->trigvar>>8)&0xff,dev->iobase+PCL711_CTR2);

		/*
		 *  Set mode to IRQ transfer
		 */

		outb(6,dev->iobase+PCL711_MODE);
		outb(0,dev->iobase+PCL711_CLRINTR);
		break;

	default:
		return -EINVAL;
	}
	
	return 0;
}
/*
 *  Sets params.
 */
static int acl8112_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || (it->pval>8 && devpriv->is_dg) || (it->pval>11 && !devpriv->is_dg) )return -EINVAL;
		devpriv->aip[it->chan]=it->pval;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			    0,it->pval);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

/* D/A conversion - Untested */
static int acl8112_da(comedi_device *dev,comedi_trig *it)
{
	if(it->data>0xfff)return -EINVAL;

	outb((it->data&0xff),dev->iobase+PCL711_DA_LO);
	outb((it->data>>8),dev->iobase+PCL711_DA_HI);
	
	return 0;
}

/* Digital port read - Untested */
static int acl8112_di(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	
	s.data=inb(dev->iobase+PCL711_DI_LO)|
		(inb(dev->iobase+PCL711_DI_HI)<<8);
	it->data=s.data;
	
	return 0;
}

/* Digital port write - Untested */
static int acl8112_do(comedi_device *dev,comedi_trig *it)
{
	int mask,data;
	
	if(it->n)mask=it->n; else mask=0xffff;
	data=(devpriv->last_do&~mask)|(it->data&mask);
	outb(data&0xff,dev->iobase+PCL711_DO_LO);
	outb((data>>8),dev->iobase+PCL711_DO_HI);
	devpriv->last_do=data;
	
	return 0;
}

/*  Initialization */
int acl8112_init(comedi_device *dev,comedi_devconfig *it)
{
	int i;
	int irq;
	bool is_dg;

	if(!strcmp("acl8112hg",it->name)) {
	        is_dg = 0;
		dev->board_name = "acl8112hg";
	} else if(!strcmp("acl8112",it->name) || !strcmp("acl8112dg",it->name)) {	      
	        is_dg = 1;
		dev->board_name = "acl8112dg";
	} else
	        return 0;

	dev->iobase=it->options[0];
	printk("comedi%d: acl8112: 0x%04x ",dev->minor,dev->iobase);
	if(check_region(dev->iobase,PCL711_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	
	dev->driver_name="acl8112";
	
	irq=it->options[1];

	if(irq<0 || irq==1 || irq>14){
		printk("irq out of range\n");
		return -EIO;
	}
	if(irq){
		if(request_irq(irq,acl8112_interrupt,SA_INTERRUPT,"acl8112",dev)){
			printk("unable to allocate irq %d\n",irq);
			irq=0;
		}else{
			printk("( irq = %d )\n",irq);
		}
	}
	dev->irq=irq;
	devpriv->is_dg = is_dg;
	
	request_region(dev->iobase,PCL711_SIZE,"acl8112");
	dev->iosize=PCL711_SIZE;

	dev->chan=20; /* 16 A/D, 2 D/A, Digital In, Digital Out */

	dev->private=kmalloc(sizeof(acl8112_private)+
		dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	if(!dev->private){
		return -ENOMEM;
	}
	memset(dev->private,0,sizeof(acl8112_private)+
		dev->chan*sizeof(comedi_chinfo));
	dev->chinfo=dev->private+sizeof(acl8112_private);

	/*
	 *  A/D channels
	 */
	for(i=0;i<16;i++){
		dev->chinfo[i].itrig=acl8112_iai;
		dev->chinfo[i].trig=acl8112_ai;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);

		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
			 COMEDI_RDONLY, (is_dg) ? RANGE_acl8112dg_ai : RANGE_acl8112hg_ai);
		dev->chinfo[i].sp=acl8112_aisp;
	}

	/* 
	 *  ACL8112 has two 12-bit, monolithic D/A channels 
	 */
	for(i=16; i<18; i++) {
	        dev->chinfo[i].itrig=acl8112_da;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			 COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			 COMEDI_IO_OUTPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
			 COMEDI_RDONLY,RANGE_bipolar5);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}

	/*
	 *  16-bit digital in
	 */
	dev->chinfo[i].itrig=acl8112_di;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_INPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;

	/*
	 *  16-bit digital out
	 */

	dev->chinfo[i].itrig=acl8112_do;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;
	
	comedi_initbuf(&dev->rdbuf);

	dev->rem=acl8112_rem;

	/* clear DAC */
	outb(0,dev->iobase+PCL711_DA_LO);
	outb(0,dev->iobase+PCL711_DA_HI);

	printk("\n");
	
	return 1;
}

/*
 *  Removes device
 */

static int acl8112_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: acl8112: remove\n",dev->minor);
	
	if(dev->irq){
		free_irq(dev->irq,dev);
	}
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);
	kfree(dev->private);

	release_region(dev->iobase,dev->iosize);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}



