/*
    module/pcl711.c
    hardware driver for PC-LabCard PCL-711 and AdSys ACL-8112
        and compatibles

    COMEDI - Linux Control and Measurement Device Interface
    Copyright (C) 1998 David A. Schleef <ds@stm.lbl.gov>
                       Janne Jalkanen <jalkanen@cs.hut.fi>
                       Eric Bunn <ebu@cs.hut.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/*
	Dave Andruczyk <xxx@tech.buffalostate.edu> also wrote a
	driver for the PCL-711.  I used a few ideas from his driver
	here.  His driver also has more comments, if you are
	interested in understanding how this driver works.

        The ACL-8112 driver was hacked from the sources of the PCL-711
        driver (the 744 chip used on the 8112 is almost the same as
        the 711b chip, but it has more I/O channels) by
        Janne Jalkanen (jalkanen@cs.hut.fi) and
        Erik Bunn (ebu@cs.hut.fi).  Remerged with the PCL-711 driver
	by ds.

        This driver supports both TRIGNOW and TRIGCLK,
        but does not yet support DMA transfers.  It also supports
        both high (HG) and low (DG) versions of the card, though
        the HG version has been untested.

*/

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <comedi_module.h>



#define PCL711_SIZE 16

#define PCL711_CTR0 0
#define PCL711_CTR1 1
#define PCL711_CTR2 2
#define PCL711_CTRCTL 3
#define PCL711_AD_LO 4
#define PCL711_DA0_LO 4
#define PCL711_AD_HI 5
#define PCL711_DA0_HI 5
#define PCL711_DI_LO 6
#define PCL711_DA1_LO 6
#define PCL711_DI_HI 7
#define PCL711_DA1_HI 7
#define PCL711_CLRINTR 8
#define PCL711_GAIN 9
#define PCL711_MUX 10
#define PCL711_MODE 11
#define PCL711_SOFTTRIG 12
#define PCL711_DO_LO 13
#define PCL711_DO_HI 14

typedef int bool;

/*
 * flags
 */

#define PCL711_TIMEOUT 100
#define PCL711_DRDY 0x10

comedi_devinit pcl711_init;
static int pcl711_rem(comedi_device *dev);

typedef struct{
	char *name;
	int is_pcl711b;
	int is_8112;
	int is_dg;
	int n_ranges;
	int n_aichan;
	int n_aochan;
	int maxirq;
	int ai_range_type;
}boardtype;

static boardtype boardtypes[]={
	{ "pcl711", 0, 0, 0, 5, 8, 1, 0, RANGE_bipolar5 },
	{ "pcl711b", 1, 0, 0, 5, 8, 1, 7, RANGE_pcl711b_ai },
	{ "acl8112hg", 0, 1, 0, 12, 16, 2, 15, RANGE_acl8112hg_ai },
	{ "acl8112dg", 0, 1, 1, 9, 16, 2, 15, RANGE_acl8112dg_ai },
};
#define n_boardtypes (sizeof(boardtypes)/sizeof(boardtype))

typedef struct{
	int board;
	int adchan;
	int ntrig;
	int aip[8];
	int last_do;
}pcl711_private;
#define devpriv ((pcl711_private *)dev->private)
#define this_board (boardtypes+dev->board)

static void pcl711_interrupt(int irq,void *d,struct pt_regs *regs)
{
	int lo,hi;
	comedi_device *dev=d;
	comedi_sample pt;
	
	hi=inb(dev->iobase+PCL711_AD_HI);
	lo=inb(dev->iobase+PCL711_AD_LO);
	outb(0,dev->iobase+PCL711_CLRINTR);
	
	pt.data=(hi<<8)|lo;
	pt.chan=devpriv->adchan;
	report_sample(dev,&pt);
	
	if(!(--devpriv->ntrig)){
		if(this_board->is_8112){
			outb(1,dev->iobase+PCL711_MODE);
		}else{
			outb(0,dev->iobase+PCL711_MODE);
		}
		dev->busy[0]=0;
		
		wake_up_interruptible(&dev->rdwait);
	}
}
	

static int pcl711_iai(comedi_device *dev,comedi_trig *it)
{
	int hi,lo,i,mask;

	if(dev->busy[0])return -EBUSY;
	outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);

	if(this_board->is_8112){

       /*
     	 *  Set the correct channel.  The two channel banks are switched
        *  using the mask value.
        *  NB: To use differential channels, you should use mask = 0x30,
        *  but I haven't written the support for this yet. /JJ
        */

       if( it->chan >= 8 ) mask = 0x20;
       else mask = 0x10;

       outb((mask | it->chan),dev->iobase+PCL711_MUX);

       /*
        *  Write the correct mode (software polling) and start polling by writing
        *  to the trigger register
        */

       outb(1,dev->iobase+PCL711_MODE);
	}else{
		outb(it->chan,dev->iobase+PCL711_MUX);
	}
	outb(0,dev->iobase+PCL711_SOFTTRIG);

	for(i=0;i<PCL711_TIMEOUT;i++){
		hi=inb(dev->iobase+PCL711_AD_HI);
		if( !(hi&PCL711_DRDY) )
			break;
		udelay(25);
	}
	if(i==PCL711_TIMEOUT){
		printk("comedi%d: pcl711: A/D timeout\n",dev->minor);
		return -ETIME;
	}
	lo=inb(dev->iobase+PCL711_AD_LO);

	it->data=((hi&0xf)<<8)|lo;
		
	return 0;
}

static int pcl711_ai(comedi_device *dev,comedi_trig *it)
{
	if(dev->busy[0])return -EBUSY;
	switch(it->trigsrc){
	case COMEDI_TRIGEXT:
		if(!this_board->is_pcl711b || dev->irq==0)return -EINVAL;
		
		dev->busy[0]=1;
		devpriv->ntrig=it->n;
		devpriv->adchan=it->chan;
		outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
		outb(it->chan,dev->iobase+PCL711_MUX);
		if(this_board->is_8112){
               /*
                *  Set mode to "no internal trigger"
                */
               		outb(0,dev->iobase+PCL711_MODE);
		}else{
			outb((dev->irq<<4)|3,dev->iobase+PCL711_MODE);
		}
		
		break;
	case COMEDI_TRIGCLK:
		if(!dev->irq)return -EINVAL;
		
		dev->busy[0]=1;
		devpriv->ntrig=it->n;
		devpriv->adchan=it->chan;
		outb(devpriv->aip[it->chan],dev->iobase+PCL711_GAIN);
		outb(it->chan,dev->iobase+PCL711_MUX);
		

               /*
                *  Set timers
                *  0x74 = Select Counter 1 | LSB/MSB | Mode=2 | Binary
                *        Mode 2 = Rate generator
                *
                *  0xb4 = Select Counter 2 | LSB/MSB | Mode=2 | Binary
                */






		outb(0x74,dev->iobase+PCL711_CTRCTL);
		outb((it->trigvar>>16)&0xff,dev->iobase+PCL711_CTR1);
		outb((it->trigvar>>24)&0xff,dev->iobase+PCL711_CTR1);
		outb(0xb4,dev->iobase+PCL711_CTRCTL);
		if(this_board->is_8112){
	               outb(it->trigvar&0xff,dev->iobase+PCL711_CTR2);
	               outb((it->trigvar>>8)&0xff,dev->iobase+PCL711_CTR2);
	
	               /*
	                *  Set mode to IRQ transfer
	                */
	
	               outb(6,dev->iobase+PCL711_MODE);
	               outb(0,dev->iobase+PCL711_CLRINTR);

		}else{
			outb(it->trigvar&0xff,dev->iobase+PCL711_CTR1);
			outb((it->trigvar>>8)&0xff,dev->iobase+PCL711_CTR1);
	
			outb((dev->irq<<4)|3,dev->iobase+PCL711_MODE);
		}
		
		break;
	default:
		return -EINVAL;
	}
	
	return 0;
}


/*
 *  Sets params.
 */

static int pcl711_aisp(comedi_device *dev,comedi_param *it)
{
	switch(it->pnum){
	case COMEDI_RANGE:
		if(it->pval<0 || it->pval>=this_board->n_ranges)
			return -EINVAL;
		devpriv->aip[it->chan]=it->pval;
		changeparam(&dev->chinfo[it->chan].paramlist,COMEDI_RANGE,
			0,it->pval);
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int pcl711_da(comedi_device *dev,comedi_trig *it)
{
	int 	da_chan=dev->chinfo[it->chan].data;
	
	if(it->data>0xfff)return -EINVAL;

	outb((it->data&0xff),dev->iobase+(da_chan? PCL711_DA1_LO : PCL711_DA0_LO ) );
	outb((it->data>>8),dev->iobase+(da_chan? PCL711_DA1_HI : PCL711_DA0_HI) );
	
	return 0;
}

/* Digital port read - Untested on 8112 */
static int pcl711_di(comedi_device *dev,comedi_trig *it)
{
	comedi_sample s;
	
	s.data=inb(dev->iobase+PCL711_DI_LO)|
		(inb(dev->iobase+PCL711_DI_HI)<<8);
	it->data=s.data;
	
	return 0;
}

/* Digital port write - Untested on 8112 */
static int pcl711_do(comedi_device *dev,comedi_trig *it)
{
	int mask,data;
	
	if(it->n)mask=it->n; else mask=0xffff;
	data=(devpriv->last_do&~mask)|(it->data&mask);
	outb(data&0xff,dev->iobase+PCL711_DO_LO);
	outb((data>>8),dev->iobase+PCL711_DO_HI);
	devpriv->last_do=data;
	
	return 0;
}

/*  Free any resources that we have claimed  */
static void free_resources(comedi_device *dev)
{
	if(dev->irq)
		free_irq(dev->irq,dev);
	
	if(dev->private)
		kfree(dev->private);
	
	if(dev->iobase)
		release_region(dev->iobase,dev->iosize);
	
	if(dev->chinfo)
		kfree(dev->chinfo);
	
}

/*  Initialization */
int pcl711_init(comedi_device *dev,comedi_devconfig *it)
{
	int		iobase;
	int		irq;
	int		i;
	int		board=-1;
	
	for(i=0;i<n_boardtypes;i++){
		if(!strcmp(boardtypes[i].name,it->name)){
			board=i;
			break;
		}
	}
	if(board<0)
		return 0;

	/* claim our I/O space */

	iobase=it->options[0];
	printk("comedi%d: pcl711: 0x%04x ",dev->minor,iobase);
	if(check_region(iobase,PCL711_SIZE)<0){
		printk("I/O port conflict\n");
		return -EIO;
	}
	request_region(dev->iobase,PCL711_SIZE,"pcl711");
	dev->iobase=iobase;
	dev->iosize=PCL711_SIZE;
	
	/* there should be a sanity check here */
	
	/* set up some name stuff */
	dev->board_name=boardtypes[board].name;
	dev->driver_name="pcl711";
	dev->board=board;
	
	/* grab our IRQ */
	irq=it->options[1];
	if(irq<0 || irq>boardtypes[board].maxirq){
		printk("irq out of range\n");
		free_resources(dev);
		return -EINVAL;
	}
	if(irq){
		if(request_irq(irq,pcl711_interrupt,SA_INTERRUPT,"pcl711",dev)){
			printk("unable to allocate irq %d\n",irq);
			free_resources(dev);
			return -EINVAL;
		}else{
			printk("( irq = %d )\n",irq);
		}
	}
	dev->irq=irq;
	
	dev->private=kmalloc(sizeof(pcl711_private),GFP_KERNEL);

	dev->chan=this_board->n_aichan+this_board->n_aochan+2;
	dev->chinfo=kmalloc(dev->chan*sizeof(comedi_chinfo),GFP_KERNEL);
	
	if(!dev->private || !dev->chinfo){
		printk("out of memory\n");
		free_resources(dev);
		return -ENOMEM;
	}

	memset(dev->private,0,sizeof(pcl711_private));
	memset(dev->chinfo,0,dev->chan*sizeof(comedi_chinfo));
	
	for(i=0;i<this_board->n_aichan;i++){
		dev->chinfo[i].itrig=pcl711_iai;
		dev->chinfo[i].trig=pcl711_ai;
		dev->chinfo[i].sp=NULL;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,
			12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,
			0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_INPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,
			COMEDI_RDONLY,0);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE,0,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
			COMEDI_RDONLY,this_board->ai_range_type);
		dev->chinfo[i].sp=pcl711_aisp;
	}

	for(;i<this_board->n_aichan+this_board->n_aochan;i++){
		dev->chinfo[i].itrig=pcl711_da;
		dev->chinfo[i].sp=NULL;
		dev->chinfo[i].data=i-this_board->n_aichan;
		addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,12);
		addparam(&dev->chinfo[i].paramlist,COMEDI_MAXDATA,COMEDI_RDONLY,0xfff);
		addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
			COMEDI_AD_ANALOG);
		addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
			COMEDI_IO_OUTPUT);
		addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
		addparam(&dev->chinfo[i].paramlist,COMEDI_RANGE_TYPE,
			COMEDI_RDONLY,RANGE_bipolar5);
		dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	}

        /*
         *  16-bit digital in
         */
	dev->chinfo[i].itrig=pcl711_di;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_INPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;

        /*
         *  16-bit digital out
         */

	dev->chinfo[i].itrig=pcl711_do;
	dev->chinfo[i].sp=NULL;
	addparam(&dev->chinfo[i].paramlist,COMEDI_BITS,COMEDI_RDONLY,16);
	addparam(&dev->chinfo[i].paramlist,COMEDI_AD,COMEDI_RDONLY,
		COMEDI_AD_DIGITAL);
	addparam(&dev->chinfo[i].paramlist,COMEDI_IO,COMEDI_RDONLY,
		COMEDI_IO_OUTPUT);
	addparam(&dev->chinfo[i].paramlist,COMEDI_LASTDATA,COMEDI_RDONLY,0);
	dev->chinfo[i].lastdata=dev->chinfo[i].paramlist;
	i++;
	
	comedi_initbuf(&dev->rdbuf);

	dev->rem=pcl711_rem;

	/* clear DAC */
	outb(0,dev->iobase+PCL711_DA0_LO);
	outb(0,dev->iobase+PCL711_DA0_HI);
	outb(0,dev->iobase+PCL711_DA1_LO);
	outb(0,dev->iobase+PCL711_DA1_HI);

	printk("\n");
	
	return 1;
}


/*
 *  Removes device
 */

static int pcl711_rem(comedi_device *dev)
{
	int i;
	
	printk("comedi%d: pcl711: remove\n",dev->minor);
	
	for(i=0;i<dev->chan;i++)
		delparamlist(dev->chinfo[i].paramlist);

	free_resources(dev);

	comedi_freebuf(&dev->rdbuf);

	return 0;
}



